from click.testing import CliRunner

from seeqret.db_utils import debug_fetch_users, debug_secrets
from seeqret.main import cli, user, users, init, list
from seeqret.cli_group_add import key
from tests.clirunner_utils import print_result


def test_add_key():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=test',
            '--email=test@example.com',
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0

        assert len(debug_secrets()) == 0
        result = runner.invoke(list)
        assert result.exit_code == 0

        result = runner.invoke(key, [
            'FOO', 'BAR',
            '--app=myapp',
            '--env=dev'
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
        assert len(debug_fetch_users()) == 1



from .serializer import serializer, BaseSerializer
from ..models import Secret


@serializer
class EnvSerializer(BaseSerializer):
    """Export to .env file.
    """
    version = 1
    tag = 'env'

    def dumps(self, secrets, system) -> str:
        if system == 'linux':
            return "don't write .env files on linux"

        res = []
        for secret in secrets:
            res.append(f'{secret.key}={secret.value}')

        return '\n'.join(res)

    def load(self, text: str, **kw) -> list[Secret]:
        lines = text.split('\n')
        return [Secret(
            key=line.split('=', 1)[0],
            plaintext_value=line.split('=', 1)[1].strip('"'),
            app=kw.get('app', '*'),
            env=kw.get('env', '*'),
            type=kw.get('type', 'str')
        ) for line in lines]



import os
import sqlite3

from seeqret.run_utils import get_seeqret_dir
# from os import abort

# import click


# def fetch_admin(cn):
#     admin = cn.execute('''
#         select username, email, pubkey
#         from users
#         where id = 1
#     ''').fetchone()
#     if not admin:
#         click.secho('No admin user', fg='red')
#         abort()
#     return dict(username=admin[0], email=admin[1], pubkey=admin[2])


# def fetch_user(cn, username):
#     user = cn.execute('''
#         select username, email, pubkey
#         from users
#         where username = ?
#     ''', [username]).fetchone()
#     if not user:
#         return None
#     return dict(username=user[0], email=user[1], pubkey=user[2])


def debug_fetch_users():
    cn = sqlite3.connect(os.path.join(get_seeqret_dir(), 'seeqrets.db'))
    with cn:
        return [rec[0] for rec in cn.execute("""
            select username
            from users
            order by username
        """)]


def debug_secrets():
    cn = sqlite3.connect(os.path.join(get_seeqret_dir(), 'seeqrets.db'))
    with cn:
        return [rec for rec in cn.execute("""
            select app, env, key, value
            from secrets
            order by app, env, key
        """)]



"""
Helper module to send json encoded data from Python.
(the misspelling is intentional ;-)
"""
# pylint:disable=E0202
import collections
import datetime
import decimal
import json
# import re

# # Call JSON.parse() if dk.jason.parse() is not available
# # (the re.sub() call removes all spaces, which is currently safe).
# CLIENT_PARSE_FN = re.sub(r'\s+', "", """
#     function (val) {
#         return (dk && dk.jason && dk.jason.parse) ?
#             dk.jason.parse(val) : JSON.parse(val)
#     }
# """)


# Are we sending a simple value, i.e. values that don't need the double parse
# required when sending '@type:__' encoded values?
# Currently, this only checks the top level of the value.
def _is_simpleval(val):
    if isinstance(val, (decimal.Decimal, int)):
        return True
    if isinstance(val, str) and not val.startswith('@'):
        return True
    return False


class DkJSONEncoder(json.JSONEncoder):
    """Handle special cases, like Decimal...
    """

    def default(self, o):

        if isinstance(o, decimal.Decimal):
            return float(o)
        if hasattr(o, '__json__'):
            return o.__json__()
        # if isinstance(o, set):
        #     return list(o)
        if isinstance(o, datetime.datetime):
            return f'{o.isoformat()}'
        if isinstance(o, datetime.date):
            return f'{o.isoformat()}'
        # if isinstance(o, datetime.time):
        #     return dict(hour=o.hour,
        #                 minute=o.minute,
        #                 second=o.second,
        #                 microsecond=o.microsecond,
        #                 kind="TIME")

        if hasattr(o, '__dict__'):
            return {k: v
                    for k, v in o.__dict__.items()
                    if not k.startswith('_')}

        if isinstance(o, collections.abc.Mapping):
            return dict(o)

        if isinstance(o, bytes):  # pragma: no branch
            return o.decode('u8')

        if isinstance(o, collections.abc.Iterable):
            return list(o)

        return super().default(o)


def dumps(val, indent=4, sort_keys=True, cls=DkJSONEncoder):
    """Dump json value, using our special encoder class.
    """
    return json.dumps(val, indent=indent, sort_keys=sort_keys, cls=cls)


def dump2(val, **kw):
    """Dump using a compact dump format.
    """
    kw['indent'] = kw.get('indent', None)
    kw['cls'] = kw.get('cls', DkJSONEncoder)
    kw['separators'] = kw.get('separators', (',', ':'))
    return json.dumps(val, **kw)

#
# DATETIME_RE = re.compile(r'''
#     @datetime:
#         (?P<year>\d{4})
#         -(?P<mnth>\d\d?)
#         -(?P<day>\d\d?)
#         T(?P<hr>\d\d?)
#         :(?P<min>\d\d?)
#         :(?P<sec>\d\d?)
#         (?:\.(?P<ms>\d+)Z?)?
# ''', re.VERBOSE)
#
#


def _iso_to_date(s):
    try:
        return datetime.date.fromisoformat(s)
    except ValueError:
        return None


def _iso_to_datetime(s):
    try:
        return datetime.datetime.fromisoformat(s)
    except ValueError:
        return None


def obj_decoder(pairs):
    """Reverses values created by DkJSONEncoder.
    """

    # def _get_tag(value):
    #     """Return the tag part of val, if it exists.
    #        Ie. @datetime:2021-11-15T12:15:47.1234 returns @datetime:
    #     """
    #     if isinstance(value, str) and value.startswith('@'):
    #         try:
    #             value = str(value)
    #         except UnicodeEncodeError:  # pragma: nocover
    #             return None
    #         else:
    #             if ':' not in value:
    #                 return None
    #             tag, _val = value.split(':', 1)
    #             return tag + ':'
    #     else:
    #         return None

    res = {}
    for key, val in pairs:
        if isinstance(val, str):
            if dt := _iso_to_date(val):
                val = dt
            elif dt := _iso_to_datetime(val):
                val = dt

        res[key] = val
    return res


def loads(txt, **kw):
    """Load json data from txt.
    """
    if 'cls' not in kw:
        kw['object_pairs_hook'] = kw.get('object_pairs_hook', obj_decoder)
    if isinstance(txt, bytes):
        txt = txt.decode('u8')
    return json.loads(txt, **kw)


def json_eval(txt):
    """Un-serialize json value.
    """
    return loads(txt)


def jsonname(val):
    """Convert the string in val to a valid json field name.
    """
    return val.replace('.', '_')



from rich.console import Console
from rich.table import Table
import logging
logger = logging.getLogger(__name__)


def dochelp(cls) -> str:
    doc = cls.__doc__
    if not doc:
        return ""
    return doc.splitlines()[0].strip()


def as_table(headers: str | list, rows, numbered=True):
    """Output ``rows`` as a table with ``headers``.
    """
    if isinstance(headers, str):
        headers = [h.strip() for h in headers.split(',')]
    table = Table()
    if numbered:
        table.add_column('#', justify='right')
    for header in headers:
        table.add_column(header)
    for i, data in enumerate(rows):
        logger.debug('row %d: %s', i, data)
        row = getattr(data, 'row', data)
        if numbered:
            table.add_row(str(i+1), *row)
        else:
            table.add_row(*row)

    console = Console()
    console.print(table)



import sqlite3

from seeqret.fileutils import remove_file_if_exists
from seeqret.migrations.utils import *


def test_table_exists():
    cn = sqlite3.connect('test_table_exists.db')
    with cn:
        assert not table_exists(cn, 'test')
        cn.executescript("""
            create table if not exists test (
              id integer primary key
            )
        """)
        assert table_exists(cn, 'test')
    cn.close()
    remove_file_if_exists('test_table_exists.db')


def test_index_exists():
    cn = sqlite3.connect('test_index_exists.db')
    with cn:
        assert not index_exists(cn, 'test', 'idx_test_foo')
        cn.executescript("""
            create table if not exists test (
                id integer primary key,
                foo text not null
            )
        """)
        cn.commit()
        cn.executescript("""
            create unique index if not exists idx_test_foo
            on test (foo)
        """)
        cn.commit()
        assert index_exists(cn, 'test', 'idx_test_foo')
    cn.close()
    remove_file_if_exists('test_index_exists.db')



from .serializer import BaseSerializer, serializer, ValidationError
from ..models import Secret


@serializer
class CommandSerializer(BaseSerializer):
    """Output list of commands that can be pasted into terminal.
    """
    version = 1
    tag = 'command'

    def dumps(self, secrets, system) -> str:
        res = []
        # quote = '"' if system == 'win32' else "'"
        for s in secrets:
            val = s.encrypt_to_string(self.sender_private_key,
                                      self.receiver_public_key)
            res.append(
                f'seeqret load -u{self.sender.username} -scommand '
                f'-v{self.version}::{val}'
                # f'-v{quote}{self.version}::{val}{quote}'
            )
        return '\n'.join(res)

    def _validate(self, txt: str) -> Secret:
        if not isinstance(txt, str):
            raise ValidationError('Text must be a string')
        params = txt.split(':', 7)
        if len(params) != 8:
            raise ValidationError('Text must contain 8 parameters')
        version, _, fingerprint, app, env, key, type, val = params
        # print("PARAMS:", params)
        if int(version) != self.version:
            raise ValidationError(
                f"Version mismatch, found {version} - expected {self.version}"
            )

        secret = Secret(
            app=app,
            env=env,
            key=key,
            plaintext_value=Secret.decrypt_value(  # FIXME: fails here
                val, self.sender_public_key, self.receiver_private_key
            ).decode('utf-8'),
            type=type,
        )
        if fingerprint != secret.fingerprint():
            raise ValidationError("invalid fingerprint")
        return secret

    def load(self, text: str) -> list[Secret]:
        return [self._validate(text)]



import pytest

from seeqret.seeqrypt.aes_fernet import encrypt_string, decrypt_string
from seeqret.seeqrypt.utils import format_encrypted_data, get_or_create_symetric_key
from seeqret.fileutils import remove_file_if_exists

COUNTER = 0


@pytest.fixture
def symetric_key():
    global COUNTER
    COUNTER += 1
    name = f'symetric-{COUNTER}.key'
    yield get_or_create_symetric_key(name)
    remove_file_if_exists(name)


def test_aes_fernet_roundtrip(symetric_key):
    print()
    plaintext = b"Hello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit ametHello, World! lorem ipsum dolor sit amet"
    encrypted = encrypt_string(symetric_key, plaintext)
    print("ENCRYPTED:", len(encrypted))
    print(format_encrypted_data(encrypted))
    decrypted = decrypt_string(symetric_key, encrypted)
    print("DECRYPTED:", len(decrypted), decrypted)
    assert plaintext == decrypted



import textwrap
from os import abort

import click
import os
import sys

from seeqret.migrations.initialize_database import init_db
# from seeqret.db_utils import fetch_admin
from seeqret.seeqret_add import add_user
from seeqret.seeqrypt.nacl_backend import (
    generate_private_key,
    private_key,
    public_key,
    save_public_key,
)
from seeqret.storage.sqlite_storage import SqliteStorage
from .seeqrypt.utils import generate_symetric_key

from .fileutils import is_encrypted, attrib_cmd, write_binary_file
from .run_utils import run, seeqret_dir, cd

DRIVE_TYPES = {
    0: 'Drive Unknown',
    1: 'No Root Directory',
    2: 'Drive Removable',
    3: 'Drive Fixed',
    4: 'Drive Network',
    5: 'Drive CD',
    6: 'Drive RAMdisk'
}


def _validate_vault_dir(dirname, vaultname):
    # we can't store secrets in a vcs repository!
    vcs_dirs = ['.svn', '.git', '.hg', '.bzr']
    for parent in list(dirname.parents) + [dirname]:
        for vcs in vcs_dirs:
            if parent.joinpath(vcs).exists():
                if vcs == '.git':
                    # ignore vaultname directory (add to .gitignore)
                    with open(parent.joinpath('.gitignore'), 'a') as f:
                        f.write(f'\n{vaultname}\n')
                elif vcs == '.svn':
                    # ignore vaultname directory
                    run(f"svn propset svn:ignore {vaultname} .", workdir=parent)
                elif vcs == '.hg':
                    # ignore vaultname directory (add to .hgignore)
                    with open(parent.joinpath('.hgignore'), 'a') as f:
                        f.write(f'\n{vaultname}\n')
                elif vcs == '.bzr':
                    # ignore vaultname directory (add to .bzrignore)
                    with open(parent.joinpath('.bzrignore'), 'a') as f:
                        f.write(f'\n{vaultname}\n')
                else:
                    click.echo(f'{parent} is a {vcs[1:]} repository, aborting.')
                    return False

    if sys.platform == 'win32':
        from win32 import win32file
        drive = os.path.splitdrive(os.path.abspath(dirname))[0]
        drive_type = win32file.GetDriveType(drive)
        if drive_type == 4:
            click.echo(f'{drive} is not a local drive, aborting.')
            click.echo(f'win32file.GetDriveType("{drive}")'
                       f' returned {drive_type}.')
            if not click.confirm('Do you want to continue?'):
                return False

    return True


def secrets_server_init(dirname, vault_dir, curuser, email, pubkey):
    owner = run("hostname").split('.')[0]
    print("SEQRET:SERVER_INIT")
    if not _validate_vault_dir(dirname, '.seeqret'):
        return False
    setup_vault(vault_dir, curuser, type='server')
    create_user_keys(vault_dir, owner)
    init_db(vault_dir, owner, f"{owner}@{owner}")
    with cd(os.path.join(dirname, '.seeqret')):
        add_user(pubkey, curuser, email)
    # TODO: we'll need system public/private keys so exports are from the
    #       system user...
    #       This means that each user will need to add the system user
    #       (eg. hostname) as a user in their own vault.
    # create_system_key(vault_dir)


# seeqret init
def secrets_init(dirname, user, email, pubkey=None, key=None):
    # dirname is the parent of seeqret..!
    seeqret_dir = dirname / 'seeqret'

    click.echo(textwrap.dedent(f'''
        Initializing seeqret vault for {user} in {dirname}
        by creating a new directory {seeqret_dir} and setting permissions.
    '''))

    if not _validate_vault_dir(dirname, 'seeqret'):
        return False
    setup_vault(seeqret_dir, user, type='client')
    create_user_keys(seeqret_dir, user, pubkey, key)
    init_db(seeqret_dir, user, email)


def create_system_key(vault_dir):
    with cd(vault_dir):
        if os.path.exists('seeqret.key'):
            click.secho('seeqret.key already exists', fg='green')
        else:
            click.echo('Creating seeqret.key')
            generate_symetric_key('seeqret.key')
            if os.path.exists('seeqret.key'):
                click.secho('seeqret.key created', fg='green')
            else:
                click.secho('seeqret.key creation failed', fg='red')
                abort()


def create_user_keys(vault_dir, user, pubkey=None, key=None):
    with cd(vault_dir):
        click.echo('Checking for existing user keys')
        if os.path.exists('public.key') and os.path.exists('private.key'):
            click.secho(f'User keys already exist for {user}', fg='green')
        else:
            click.echo(f'Creating keys for {user}')
            if key:
                # FIXME: use the nacl_backend functions
                #        for reading/writing keys?
                write_binary_file('private.key', key.encode('ascii'))
                pkey = private_key(key)
            else:
                pkey = generate_private_key('private.key')
            if pubkey:
                write_binary_file('public.key', pubkey.encode('ascii'))
                pubkey = public_key(pubkey)
            else:
                pubkey = save_public_key('public.key', pkey)
            click.secho(f'Keys created for {user}', fg='green')
            click.secho(f'Please publish your public key: {pubkey}', fg='blue')

        create_system_key(vault_dir)

        if os.environ.get('TESTING', "") == "TRUE":
            os.environ["SEEQRET"] = os.path.abspath(vault_dir)
        else:
            if sys.platform == 'win32':
                run(f'setx SEEQRET {os.path.abspath(vault_dir)}')

    if sys.platform == 'win32':
        click.echo("I've set the SEEQRET environment variable to the vault "
                   "directory")
        click.echo("Please close this window and open a new one to continue.")
        click.echo('or run\n\n')
        click.echo(f'    set "SEQRET={os.path.abspath(vault_dir)}"')
        click.echo('\n\nin the current window to continue here.')


# seeqret upgrade
def upgrade_db():
    with seeqret_dir():
        storage = SqliteStorage()
        admin = storage.fetch_admin()
        if not admin:
            return False
        init_db(os.environ['SEEQRET'], admin.username, admin.email)


def setup_vault(vault_dir, user, type='client'):

    if os.name == 'nt':
        if not vault_dir.exists():
            click.echo(f'creating {vault_dir}.')
            vault_dir.mkdir(0o770)

        with cd(vault_dir.parent):
            seeqret_dir = str(vault_dir)
            if len(run(f"icacls {seeqret_dir}").splitlines()) >= 4:
                click.echo(f"Tightening permissions on {vault_dir}")
                click.echo("Granting (F)ull rights to current user only")
                userdomain = os.environ['USERDOMAIN']
                username = os.environ['USERNAME']
                current_user = f'{userdomain}\\{username}'
                run(f"icacls {seeqret_dir} /grant {current_user}:(F)")
                click.echo("Removing all inherited permissions")
                run(f"icacls {seeqret_dir} /inheritance:r")
                click.echo("Verifying permissions..")
                if len(run("icacls " + seeqret_dir).splitlines()) >= 4:
                    click.echo("Could not change permissions on vault_dir")
            click.echo("vault_dir permissions are ok")

            click.echo("Checking if vault_dir is indexed by windows search")
            if 'I' not in attrib_cmd(seeqret_dir):
                click.echo(f"Removing {vault_dir} from windows indexing.")
                attrib_cmd(seeqret_dir, '+I')
            else:
                click.secho(f"{vault_dir} is not indexed", fg='green')

            if not is_encrypted("seeqret"):
                click.echo(f"encrypting {vault_dir}")
                run("cipher /e seeqret")
                click.echo("Checking if encryption worked..")
                if not is_encrypted("seeqret"):
                    click.echo(
                        "cipher /e seeqret (this is very bad, aborting...)"
                    )
                    # this is very, very bad..
                    abort()
                else:
                    click.echo("vault is encrypted")
            else:
                click.echo("vault is encrypted")
    else:
        # linux... (server vault or CI)
        if type == 'client':
            vault_dir.mkdir(0o770)
            return

        click.echo(textwrap.dedent(f"""\
            - I will now create a group called 'seeqret' and add the current user to it.
            - I will also set the permissions on the {vault_dir.parent} to be
              owned by group 'seeqret' and set the groupid and sticky bits.
            - I will then set the permissions on the vault_dir to 770.
        """))

        if not click.confirm("Do you want to continue?"):
            abort()
        if not vault_dir.exists():
            click.secho('adding group seeqret...', fg='blue')
            run("sudo getent group seeqret || sudo groupadd seeqret")
            if not run("sudo getent group seeqret"):
                click.echo("Could not create group seeqret")
                abort()

            click.secho(f'adding user {user} to group seeqret...', fg='blue')
            run(f"sudo usermod -aG seeqret {user}")
            if 'seeqret' not in run(f"groups {user}"):
                click.echo("Could not add user to group seeqret")
                abort()

            click.secho(f'changing group ownership of {vault_dir.parent} to seeqret...', fg='blue')
            run(f"sudo chgrp seeqret {vault_dir.parent}")

            click.secho(f'setting permissions on {vault_dir.parent}...', fg='blue')
            run(f"sudo chmod g+s {vault_dir.parent}")
            run(f"sudo chmod g+w {vault_dir.parent}")
            try:
                vault_dir.mkdir(0o770)
            except PermissionError:
                click.echo("You must login again, and reissue the command to continue.")
                abort()
        run(f"sudo chmod g+w {vault_dir}")



from seeqret.fileutils import *


def test_appdata_dir():
    assert local_appdata_dir()
    assert roaming_appdata_dir()


def test_write_file():
    write_file('test-write-file.txt', """\
        hello world
    """)
    assert read_file('test-write-file.txt') == 'hello world\n'
    remove_file_if_exists('test-write-file.txt')


def test_read_json():
    write_file('test-read-json.json', """
        {
            "hello": "world"
        }
    """)
    assert read_json('test-read-json.json') == {"hello": "world"}
    remove_file_if_exists('test-read-json.json')



from ..models import jason
from ..models.jason import loads
from .serializer import BaseSerializer, ValidationError, serializer
from ..models.secret import hash_secrets, Secret


@serializer
class InsecureJsonSerializer(BaseSerializer):
    """For debugging (and maybe backup?)
    """
    version = 1
    tag = 'backup'

    def to_json_object(self, secrets):
        res = {
            "version": self.version,
            "from": self.sender,
            "to": self.receiver,
            "secrets": [s.to_plaintext_dict() for s in secrets],
            "signature": hash_secrets(secrets),
        }
        return res

    def dumps(self, secrets, system):
        return jason.dumps(self.to_json_object(secrets))

    def deserialize_version(self, text):
        tmp = loads(text)
        if tmp["version"] != self.version:
            # XXX: for now, bail if incorrect version
            raise ValidationError("Incorrect version")
        return [Secret(**s) for s in tmp["secrets"]]



"""
Safely transferring code secrets.
"""

import setuptools

version = '0.1.5'

setuptools.setup(
    name='seeqret',
    description='Safely transferring code secrets',
    keywords='secrets, gpg, pgp',
    version=version,
    author='Bjorn',
    url='https://github.com/thebjorn/seeqret.git',
    install_requires=[
        'rsa',
        'python-gnupg',
        'cryptography',
        'pynacl',
        'Click',
        'rich',
        'requests',
        "pywin32  ; sys_platform == 'win32'"
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    ],
    entry_points={
        'console_scripts': [
            'seeqret=seeqret.main:cli',
        ],
    },
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    include_package_data=True,
    zip_safe=False,
)
